{% extends 'allianceauth/base.html' %}
{% load i18n %}
{% load static %}

{% block page_title %}{% translate app_title %}{% endblock %}
{% block content %}
    <br />
    {% include 'standingssync/partial/menu.html' %}
    <h2>{{ page_title }}</h2>
    {% block details %}{% endblock %}
{% endblock %}

{% block extra_css %}
    <link href="{% static 'standingssync/base.css' %}" type="text/css" rel="stylesheet">
{% endblock %}
