{% extends "admin/change_list.html" %}
{% load i18n %}

{% block object-tools-items %}
    {{ block.super }}
    <li>
        <a href="{% url 'standingssync:admin_update_wars' %}" class="btn btn-high">
            {% translate "Update now" %}
        </a>
    </li>
{% endblock %}
