# Generated by Django 4.0.9 on 2023-02-10 13:25

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("eveonline", "0015_factions"),
        ("authentication", "0019_merge_20211026_0919"),
        ("standingssync", "0001_initial_new_2"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="syncedcharacter",
            name="last_error",
        ),
        migrations.RemoveField(
            model_name="syncedcharacter",
            name="last_sync",
        ),
        migrations.RemoveField(
            model_name="syncedcharacter",
            name="version_hash",
        ),
        migrations.RemoveField(
            model_name="syncmanager",
            name="last_error",
        ),
        migrations.RemoveField(
            model_name="syncmanager",
            name="last_sync",
        ),
        migrations.AddField(
            model_name="syncedcharacter",
            name="last_sync_at",
            field=models.DateTimeField(
                default=None,
                help_text="When the last successful sync was completed.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="syncmanager",
            name="last_sync_at",
            field=models.DateTimeField(
                default=None,
                help_text="When the last successful sync was completed.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="syncedcharacter",
            name="character_ownership",
            field=models.OneToOneField(
                help_text="Character to sync",
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                serialize=False,
                to="authentication.characterownership",
            ),
        ),
        migrations.AlterField(
            model_name="syncedcharacter",
            name="has_war_targets_label",
            field=models.BooleanField(
                default=None,
                help_text="Whether this character has the war target label.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="alliance",
            field=models.OneToOneField(
                help_text="Alliance to sync contacts for",
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                related_name="+",
                serialize=False,
                to="eveonline.eveallianceinfo",
            ),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="character_ownership",
            field=models.OneToOneField(
                default=None,
                help_text="alliance contacts are fetched from this character",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="authentication.characterownership",
            ),
        ),
        migrations.AlterField(
            model_name="syncmanager",
            name="version_hash",
            field=models.CharField(
                default="",
                help_text="hash over all contacts to identify when it has changed",
                max_length=32,
            ),
        ),
    ]
