{% load static %}
{% load i18n %}
{% load navactive %}

<nav class="navbar navbar-default">
    <div class="container-fluid">
        <div class="navbar-header">
            <button type="button" class="navbar-toggle collapsed" data-toggle="collapse" data-target="#bs-example-navbar-collapse-1" aria-expanded="false">
                <span class="sr-only">{% trans "Toggle navigation" %}</span>
                <span class="icon-bar"></span>
                <span class="icon-bar"></span>
                <span class="icon-bar"></span>
            </button>
            <a class="navbar-brand" href="{% url 'standingssync:characters' %}">{{ app_title }}</a>
        </div>
        <div class="collapse navbar-collapse" id="bs-example-navbar-collapse-1">
            <ul class="nav navbar-nav">
                <li class="{% navactive request 'standingssync:characters' %}">
                    <a href="{% url 'standingssync:characters' %}">My Characters</a>
                </li>
                {% if STANDINGSSYNC_ADD_WAR_TARGETS %}
                    <li class="{% navactive request 'standingssync:wars' %}">
                        <a href="{% url 'standingssync:wars' %}">
                            Current Wars
                        </a>
                    </li>
                {% endif %}
            </ul>

            <ul class="nav navbar-nav navbar-right">

                {% if perms.standingssync.add_syncmanager %}
                    <a class="btn btn-default navbar-btn" href="{% url 'standingssync:add_alliance_manager' %}">
                        Set Alliance Character
                    </a>
                {% endif %}
            </ul>
        </div>
    </div>
</nav>
